from snowddl.blueprint import ProcedureBlueprint, Ident, SchemaObjectIdentWithArgs, NameWithType, DataType, StageWithPath, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


procedure_json_schema = {
    "type": "object",
    "properties": {
        "language": {
            "type": "string"
        },
        "body": {
            "type": "string"
        },
        "arguments": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "returns": {
            "anyOf": [
                {
                    "type": "string"
                },
                {
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                }
            ]
        },
        "is_strict": {
            "type": "boolean"
        },
        "is_execute_as_caller": {
            "type": "boolean"
        },
        "runtime_version": {
            "type": "string"
        },
        "imports": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "stage": {
                        "type": "string"
                    },
                    "path": {
                        "type": "string"
                    }
                },
                "required": ["stage"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "packages": {
            "type": "array",
            "items": {
                "type": "string"
            },
            "minItems": 1
        },
        "handler": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "anyOf": [
        {
            "required": ["body", "returns"]
        },
        {
            "required": ["handler", "returns"]
        }
    ],
    "additionalProperties": False
}


class ProcedureParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("procedure", procedure_json_schema, self.process_procedure)

    def process_procedure(self, f: ParsedFile):
        arguments = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params.get('arguments', {}).items()]

        if isinstance(f.params['returns'], dict):
            # Returns table
            returns = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params['returns'].items()]
        else:
            # Returns value
            returns = DataType(f.params['returns'])

        if f.params.get('imports'):
            imports = [StageWithPath(stage_name=build_schema_object_ident(self.env_prefix, i['stage'], f.database, f.schema), path=i['path']) for i in f.params.get('imports')]
        else:
            imports = None

        if f.params.get('packages'):
            packages = f.params.get('packages')
        else:
            packages = None

        base_name = self.validate_name_with_args(f.path, arguments)

        bp = ProcedureBlueprint(
            full_name=SchemaObjectIdentWithArgs(self.env_prefix, f.database, f.schema, base_name, [a.type.base_type for a in arguments]),
            language=f.params.get('language', 'SQL'),
            body=f.params.get('body'),
            arguments=arguments,
            returns=returns,
            is_strict=f.params.get('is_strict', False),
            is_immutable=f.params.get('is_immutable', False),
            is_execute_as_caller=f.params.get('is_execute_as_caller', False),
            runtime_version=f.params.get('runtime_version'),
            imports=imports,
            packages=packages,
            handler=f.params.get('handler'),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import OutboundShareBlueprint, Grant, build_grant_name_ident_snowflake
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class OutboundShareResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.SHARE

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW SHARES")

        for r in cur:
            if r['kind'] != 'OUTBOUND':
                continue

            if r['owner'] != self.engine.context.current_role:
                continue

            # Remove organization and account prefix, shares are referred in SQL by name only
            full_name = r['name'].split('.')[-1]

            existing_objects[full_name] = {
                "share": full_name,
                "database": r['database_name'],
                "accounts": r['to'].split(',') if r['to'] else [],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(OutboundShareBlueprint)

    def create_object(self, bp: OutboundShareBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE SHARE {full_name:i}", {
            "full_name": bp.full_name,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)

        for bp_grant in bp.grants:
            self.create_grant(bp.full_name, bp_grant)

        if bp.accounts:
            self.set_accounts(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: OutboundShareBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        existing_grants = self.get_existing_share_grants(bp.full_name)

        for bp_grant in bp.grants:
            if bp_grant not in existing_grants:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for ex_grant in existing_grants:
            if ex_grant not in bp.grants:
                self.drop_grant(bp.full_name, ex_grant)
                result = ResolveResult.GRANT

        # SHOW SHARES command returns only 3 accounts for each OUTBOUND SHARE
        # If you have to set more accounts, it will work, but SnowDDL will be forced to run SET ACCOUNTS every time
        if [str(a) for a in bp.accounts] != row['accounts']:
            self.set_accounts(bp)
            result = ResolveResult.ALTER

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER SHARE {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            }, condition=self.engine.settings.execute_outbound_share)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP SHARE {share_name:i}", {
            "share_name": row['share'],
        }, condition=self.engine.settings.execute_outbound_share)

        return ResolveResult.DROP

    def set_accounts(self, bp: OutboundShareBlueprint):
        if bp.accounts:
            query = self.engine.query_builder()

            query.append("ALTER SHARE {full_name:i} SET ACCOUNTS = {accounts:i}", {
                "full_name": bp.full_name,
                "accounts": bp.accounts,
            })

            if bp.share_restrictions is not None:
                query.append_nl("SHARE_RESTRICTIONS = {share_restrictions:b}", {
                    "share_restrictions": bp.share_restrictions
                })

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_outbound_share)
        else:
            # Special case for all consumer accounts being removed from SHARE
            self.engine.execute_unsafe_ddl("ALTER SHARE {full_name:i} SET ACCOUNTS = NULL", {
                "full_name": bp.full_name,
            }, condition=self.engine.settings.execute_outbound_share)

    def create_grant(self, share_name, grant: Grant):
        self.engine.execute_unsafe_ddl("GRANT {privilege:r} ON {on:r} {name:i} TO SHARE {share_name:i}", {
            "privilege": grant.privilege,
            "on": grant.on.singular,
            "name": grant.name,
            "share_name": share_name,
        }, condition=self.engine.settings.execute_outbound_share)

    def drop_grant(self, share_name, grant: Grant):
        self.engine.execute_unsafe_ddl("REVOKE {privilege:r} ON {on:r} {name:i} FROM SHARE {share_name:i}", {
            "privilege": grant.privilege,
            "on": grant.on.singular,
            "name": grant.name,
            "share_name": share_name,
        }, condition=self.engine.settings.execute_outbound_share)

    def get_existing_share_grants(self, share_name):
        grants = []

        cur = self.engine.execute_meta("SHOW GRANTS TO SHARE {share_name:i}", {
            "share_name": share_name,
        })

        for r in cur:
            grants.append(Grant(
                privilege=r['privilege'],
                on=ObjectType[r['granted_on']],
                name=build_grant_name_ident_snowflake(r['name'], ObjectType[r['granted_on']]),
            ))

        return grants



from snowddl.blueprint import NetworkPolicyBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class NetworkPolicyResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.NETWORK_POLICY

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW NETWORK POLICIES")

        for r in cur:
            # SHOW NETWORK POLICIES does not support LIKE filter, so it has to be applied manually in the code
            if self.config.env_prefix and not str(r['name']).startswith(self.config.env_prefix):
                continue

            # SHOW NETWORK POLICIES does not return "owner", so additional ownership check cannot be performed

            existing_objects[r['name']] = {
                "name": r['name'],
                "entries_in_allowed_ip_list": r['entries_in_allowed_ip_list'],
                "entries_in_blocked_ip_list": r['entries_in_blocked_ip_list'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(NetworkPolicyBlueprint)

    def create_object(self, bp: NetworkPolicyBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE NETWORK POLICY {name:i}", {
            "name": bp.full_name,
        })

        query.append_nl("ALLOWED_IP_LIST = ({allowed_ip_list})", {
            "allowed_ip_list": bp.allowed_ip_list,
        })

        if bp.blocked_ip_list:
            query.append_nl("BLOCKED_IP_LIST = ({blocked_ip_list})", {
                "blocked_ip_list": bp.blocked_ip_list,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_network_policy)

        return ResolveResult.CREATE

    def compare_object(self, bp: NetworkPolicyBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        cur = self.engine.execute_meta("DESC NETWORK POLICY {name:i}", {
            "name": bp.full_name,
        })

        existing_allowed_ip_list = []
        existing_blocked_ip_list = []

        for r in cur:
            if r['name'] == 'ALLOWED_IP_LIST':
                existing_allowed_ip_list = str(r['value']).split(',')
            elif r['name'] == 'BLOCKED_IP_LIST':
                existing_blocked_ip_list = str(r['value']).split(',')

        if sorted(bp.allowed_ip_list) != sorted(existing_allowed_ip_list):
            self.engine.execute_unsafe_ddl("ALTER NETWORK POLICY {name:i} SET ALLOWED_IP_LIST = ({allowed_ip_list})", {
                "name": bp.full_name,
                "allowed_ip_list": bp.allowed_ip_list,
            }, condition=self.engine.settings.execute_network_policy)

            result = ResolveResult.ALTER

        if sorted(bp.blocked_ip_list) != sorted(existing_blocked_ip_list):
            if bp.blocked_ip_list:
                self.engine.execute_unsafe_ddl("ALTER NETWORK POLICY {name:i} SET BLOCKED_IP_LIST = ({blocked_ip_list})", {
                    "name": bp.full_name,
                    "blocked_ip_list": bp.blocked_ip_list,
                }, condition=self.engine.settings.execute_network_policy)
            else:
                self.engine.execute_unsafe_ddl("ALTER NETWORK POLICY {name:i} SET BLOCKED_IP_LIST = ()", {
                    "name": bp.full_name,
                }, condition=self.engine.settings.execute_network_policy)

            result = ResolveResult.ALTER

        if bp.comment != row['comment']:
            self.engine.execute_unsafe_ddl("ALTER NETWORK POLICY {name:i} SET COMMENT = {comment}", {
                "name": bp.full_name,
                "comment": bp.comment if bp.comment else None,
            }, condition=self.engine.settings.execute_network_policy)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP NETWORK POLICY {name:i}", {
            "name": row['name'],
        }, condition=self.engine.settings.execute_network_policy)

        return ResolveResult.DROP



from snowddl.blueprint import UserBlueprint, AccountObjectIdent, build_role_ident, build_default_namespace_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.business_role import business_role_json_schema


user_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "login_name": {
                "type": "string"
            },
            "display_name": {
                "type": "string"
            },
            "first_name": {
                "type": "string"
            },
            "last_name": {
                "type": "string"
            },
            "email": {
                "type": "string"
            },
            "disabled": {
                "type": "boolean"
            },
            "password": {
                "type": "string"
            },
            "rsa_public_key": {
                "type": "string"
            },
            "rsa_public_key_2": {
                "type": "string"
            },
            "default_warehouse": {
                "type": "string"
            },
            "default_namespace": {
                "type": "string"
            },
            "session_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "business_roles": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 1
            },
            "comment": {
                "type": "string"
            }
        },
        "additionalProperties": False
    }
}


class UserParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'user.yaml', user_json_schema, self.process_user)

    def process_user(self, f: ParsedFile):
        default_warehouse_map = self.get_default_warehouse_map()

        for user_name, user in f.params.items():
            business_roles = []
            default_warehouse = user.get('default_warehouse')

            for business_role_name in user['business_roles']:
                business_roles.append(build_role_ident(self.env_prefix, business_role_name, self.config.BUSINESS_ROLE_SUFFIX))

                if default_warehouse is None:
                    default_warehouse = default_warehouse_map.get(business_role_name)

            full_user_name = AccountObjectIdent(self.env_prefix, user_name)

            if user.get('login_name'):
                # Login name is string, not identifier (!), special characters like '@#!' are permitted
                # But it still requires env_prefix due to unique constraint
                login_name = f"{self.config.env_prefix}{user.get('login_name')}".upper()
            else:
                login_name = str(full_user_name)

            if user.get('display_name'):
                display_name = str(user.get('display_name')).upper()
            else:
                display_name = str(full_user_name)

            bp = UserBlueprint(
                full_name=full_user_name,
                login_name=login_name,
                display_name=display_name,
                first_name=user.get('first_name'),
                last_name=user.get('last_name'),
                email=user.get('email'),
                disabled=user.get('disabled', False),
                password=user.get('password'),
                rsa_public_key=user.get('rsa_public_key').replace(' ', '') if user.get('rsa_public_key') else None,
                rsa_public_key_2=user.get('rsa_public_key_2').replace(' ', '') if user.get('rsa_public_key_2') else None,
                default_warehouse=AccountObjectIdent(self.env_prefix, default_warehouse) if default_warehouse else None,
                default_namespace=build_default_namespace_ident(self.env_prefix, user.get('default_namespace')) if user.get('default_namespace') else None,
                session_params=self.normalise_params_dict(user.get('session_params', {})),
                business_roles=business_roles,
                comment=user.get('comment'),
            )

            self.config.add_blueprint(bp)

    def get_default_warehouse_map(self):
        default_warehouse_map = {}

        business_role_config = self.parse_single_file(self.base_path / 'business_role.yaml', business_role_json_schema)

        for business_role_name, business_role in business_role_config.items():
            if 'warehouse_usage' not in business_role:
                continue

            default_warehouse_map[business_role_name] = business_role['warehouse_usage'][0]

        return default_warehouse_map



from .blueprint import (
    AbstractBlueprint,
    SchemaObjectBlueprint,
    DependsOnMixin,
    RoleBlueprint,
    AccountParameterBlueprint,
    BusinessRoleBlueprint,
    DatabaseBlueprint,
    DatabaseShareBlueprint,
    ExternalFunctionBlueprint,
    ExternalTableBlueprint,
    FileFormatBlueprint,
    ForeignKeyBlueprint,
    FunctionBlueprint,
    MaterializedViewBlueprint,
    MaskingPolicyBlueprint,
    NetworkPolicyBlueprint,
    OutboundShareBlueprint,
    PipeBlueprint,
    PrimaryKeyBlueprint,
    ProcedureBlueprint,
    ResourceMonitorBlueprint,
    RowAccessPolicyBlueprint,
    SchemaBlueprint,
    SchemaRoleBlueprint,
    StageBlueprint,
    StageFileBlueprint,
    SequenceBlueprint,
    StreamBlueprint,
    TableBlueprint,
    TagBlueprint,
    TaskBlueprint,
    TechRoleBlueprint,
    UniqueKeyBlueprint,
    UserBlueprint,
    ViewBlueprint,
    WarehouseBlueprint,
    T_Blueprint,
)

from .column import ExternalTableColumn, TableColumn, ViewColumn, NameWithType, SearchOptimizationItem
from .data_type import BaseDataType, DataType
from .edition import Edition
from .grant import Grant, FutureGrant

from .ident import (
    AbstractIdent,
    AbstractIdentWithPrefix,
    Ident,
    AccountIdent,
    AccountObjectIdent,
    DatabaseIdent,
    InboundShareIdent,
    OutboundShareIdent,
    SchemaIdent,
    SchemaObjectIdent,
    SchemaObjectIdentWithArgs,
    StageFileIdent,
    TableConstraintIdent,
)

from .ident_builder import (
    build_schema_object_ident,
    build_role_ident,
    build_grant_name_ident_snowflake,
    build_default_namespace_ident,
)

from .object_type import ObjectType
from .reference import MaskingPolicyReference, RowAccessPolicyReference, TagReference
from .stage import StageWithPath, StageUploadFile



from snowddl.blueprint import ObjectType
from snowddl.converter.abc_converter import AbstractConverter, ConvertResult
from snowddl.parser.database import database_json_schema


class DatabaseConverter(AbstractConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.DATABASE

    def get_existing_objects(self):
        return self.engine.schema_cache.databases

    def dump_object(self, row):
        data = {}

        if row['is_transient']:
            data['is_transient'] = True

        if row['comment']:
            data['comment'] = row['comment']

        object_path = self.base_path / self._normalise_name_with_prefix(row['database'])
        object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

        self._dump_file(object_path / 'params.yaml', data, database_json_schema)
        return ConvertResult.DUMP



from typing import Union


def coalesce(val, default=None):
    return default if val is None else val


def compare_dynamic_param_value(bp_value: Union[bool,int,float,str], existing_value: str) -> bool:
    if isinstance(bp_value, bool):
        if bp_value == True and existing_value == 'true':
            return True

        if bp_value == False and existing_value == 'false':
            return True

    elif isinstance(bp_value, int):
        if bp_value == int(existing_value):
            return True

    elif isinstance(bp_value, float):
        if bp_value == float(existing_value):
            return True

    else:
        if str(bp_value) == existing_value:
            return True

    return False


def dtypes_from_arguments(arguments: str) -> str:
    start_dtypes_idx = arguments.index('(')
    finish_dtypes_idx = arguments.index(')')

    return ','.join([a.strip(' ') for a in arguments[start_dtypes_idx+1:finish_dtypes_idx].split(',')])



from collections import defaultdict
from dataclasses import fields
from fnmatch import translate
from pathlib import Path
from re import compile
from typing import List, Dict, Type, Union

from snowddl.blueprint import AbstractBlueprint, AbstractIdentWithPrefix, T_Blueprint


class SnowDDLConfig:
    BUSINESS_ROLE_SUFFIX = 'B_ROLE'
    INBOUND_SHARE_ROLE_SUFFIX = 'IS_ROLE'
    SCHEMA_ROLE_SUFFIX = 'S_ROLE'
    TECH_ROLE_SUFFIX = 'T_ROLE'
    USER_ROLE_SUFFIX = 'U_ROLE'
    WAREHOUSE_ROLE_SUFFIX = 'W_ROLE'

    def __init__(self, env_prefix=None):
        self.env_prefix = self._init_env_prefix(env_prefix)

        self.blueprints: Dict[Type[T_Blueprint], Dict[str,T_Blueprint]] = defaultdict(dict)
        self.errors: List[dict] = []

        self.placeholders: Dict[str,Union[bool,float,int,str]] = {}

    def get_blueprints_by_type(self, cls: Type[T_Blueprint]) -> Dict[str,T_Blueprint]:
        return self.blueprints.get(cls, {})

    def get_blueprints_by_type_and_pattern(self, cls: Type[T_Blueprint], pattern: str) -> Dict[str,T_Blueprint]:
        pattern = pattern.upper()
        type_blueprints = self.blueprints.get(cls, {})

        # Add env prefix to pattern IF blueprint supports it
        for f in fields(cls):
            if f.name == 'full_name':
                if issubclass(f.type, AbstractIdentWithPrefix):
                    pattern = f"{self.env_prefix}{pattern}"

            break

        # Use Unix-style wildcards if any special characters detected in pattern
        if any(special_char in pattern for special_char in ['*', '?', '[', ']']):
            regexp = compile(translate(pattern))

            return {full_name: bp for full_name, bp in type_blueprints.items() if regexp.match(full_name)}

        # Use basic key matching otherwise
        if pattern in type_blueprints:
            return {pattern: type_blueprints[pattern]}

        return {}

    def get_placeholder(self, name: str) -> Union[bool,float,int,str]:
        if name not in self.placeholders:
            raise ValueError(f"Unknown placeholder [{name}]")

        return self.placeholders[name]

    def add_blueprint(self, bp: AbstractBlueprint):
        self.blueprints[bp.__class__][str(bp.full_name)] = bp

    def add_error(self, path: Path, e: Exception):
        self.errors.append({
            "path": path,
            "error": e,
        })

    def add_placeholder(self, name: str, value: Union[bool,float,int,str]):
        self.placeholders[name] = value

    def _init_env_prefix(self, env_prefix):
        if env_prefix:
            env_prefix = str(env_prefix).upper()

            if '__' in env_prefix:
                raise ValueError(f"Env prefix [{env_prefix}] cannot contain [__] double underscore")

            if env_prefix.endswith('_'):
                raise ValueError(f"Env prefix [{env_prefix}] cannot end with [_] underscore")

            return f"{env_prefix}__"

        return ''



from snowddl.blueprint import AccountParameterBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import compare_dynamic_param_value


class AccountParameterResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.ACCOUNT_PARAMETER

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW PARAMETERS FOR ACCOUNT")

        for r in cur:
            existing_objects[r['key']] = {
                "key": r['key'],
                "value": r['value'],
                "default": r['default'],
                "level": r['level'],
                "type": r['type'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(AccountParameterBlueprint)

    def create_object(self, bp: AccountParameterBlueprint):
        raise ValueError(f"Unknown account parameter [{bp.full_name}]")

    def compare_object(self, bp: AccountParameterBlueprint, row: dict):
        if compare_dynamic_param_value(bp.value, row['value']):
            return ResolveResult.NOCHANGE

        self.engine.execute_unsafe_ddl("ALTER ACCOUNT SET {param_name:r} = {param_value:dp}", {
            "param_name": row['key'],
            "param_value": bp.value,
        }, condition=self.engine.settings.execute_account_params)

        return ResolveResult.ALTER

    def drop_object(self, row: dict):
        if row['value'] != row['default'] and row['level'] == 'ACCOUNT':
            self.engine.execute_unsafe_ddl("ALTER ACCOUNT UNSET {param_name:r}", {
                "param_name": row['key'],
            }, condition=self.engine.settings.execute_account_params)

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def destroy(self):
        pass



from snowddl.blueprint import DatabaseIdent, DatabaseShareBlueprint, InboundShareIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


inbound_share_json_schema = {
    "type": "object",
    "additionalProperties": {
        "properties": {
            "share_name": {
                "type": "string"
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["share_name"],
        "additionalProperties": False
    }
}


class InboundShareParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'inbound_share.yaml', inbound_share_json_schema, self.process_inbound_share)

    def process_inbound_share(self, f: ParsedFile):
        for database_name, share in f.params.items():
            share_name_parts = share['share_name'].split('.')

            if len(share_name_parts) != 3:
                raise ValueError(f"Invalid inbound share name format [{share['share_name']}], expected identifier with exactly 3 parts: <organization>.<account>.<share>")

            bp = DatabaseShareBlueprint(
                # Inbound shares do not support env_prefix
                full_name=DatabaseIdent('', database_name),
                share_name=InboundShareIdent(*share_name_parts),
                comment=share.get('comment'),
            )

            self.config.add_blueprint(bp)



from yaml import SafeDumper, add_representer

class SnowDDLDumper(SafeDumper):
    pass

class YamlFoldedStr(str):
    pass

class YamlLiteralStr(str):
    pass

def folded_str_presenter(dumper: SnowDDLDumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=">")

def literal_str_presenter(dumper: SnowDDLDumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")

add_representer(YamlFoldedStr, folded_str_presenter, Dumper=SnowDDLDumper)
add_representer(YamlLiteralStr, literal_str_presenter, Dumper=SnowDDLDumper)



from snowddl.blueprint import ExternalTableBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class ExternalTableResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.EXTERNAL_TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW EXTERNAL TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "invalid": r['invalid'] == "true",
                "invalid_reason": r['invalid_reason'],
                "stage": r['stage'],
                "location": r['location'],
                "file_format_name": r['file_format_name'],
                "file_format_type": r['file_format_type'],
                "notification_channel": r['notification_channel'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ExternalTableBlueprint)

    def create_object(self, bp: ExternalTableBlueprint):
        query = self._build_create_external_table(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON TABLE {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: ExternalTableBlueprint, row: dict):
        query = self._build_create_external_table(bp)

        if not query.compare_short_hash(row['comment']) or row['invalid']:
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON TABLE {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP EXTERNAL TABLE {database:i}.{schema:i}.{table_name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "table_name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_external_table(self, bp: ExternalTableBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE EXTERNAL TABLE {full_name:i}", {
            "full_name": bp.full_name,
        })

        if bp.columns:
            query.append_nl("(")

            for idx, c in enumerate(bp.columns):
                query.append_nl("    {comma:r}{col_name:i} {col_type:r} AS ({col_expr:r})", {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": c.name,
                    "col_type": c.type,
                    "col_expr": c.expr,
                })

                if c.not_null:
                    query.append("NOT NULL")

                if c.comment:
                    query.append("COMMENT {comment}", {
                        "comment": c.comment,
                    })

            query.append_nl(")")

        if bp.partition_by:
            query.append_nl("PARTITION BY ({partition_by:i})", {
                "partition_by": bp.partition_by,
            })

        if bp.partition_type:
            query.append_nl("PARTITION_TYPE = {partition_type}", {
                "partition_type": bp.partition_type,
            })

        query.append_nl("LOCATION = @{stage_name:i}{path:r}", {
            "stage_name": bp.location_stage,
            "path": coalesce(bp.location_path, ''),
        })

        if bp.location_pattern:
            query.append_nl("PATTERN = {pattern}", {
                "pattern": bp.location_pattern,
            })

        query.append_nl("FILE_FORMAT = (FORMAT_NAME = {file_format:i})", {
            "file_format": bp.file_format,
        })

        if bp.refresh_on_create is not None:
            query.append_nl("REFRESH_ON_CREATE = {refresh_on_create:b}", {
                "refresh_on_create": bp.refresh_on_create,
            })

        if bp.auto_refresh is not None:
            query.append_nl("AUTO_REFRESH = {auto_refresh:b}", {
                "auto_refresh": bp.auto_refresh,
            })

        if bp.aws_sns_topic:
            query.append_nl("AWS_SNS_TOPIC = {aws_sns_topic}", {
                "aws_sns_topic": bp.aws_sns_topic,
            })

        if bp.table_format:
            query.append_nl("TABLE_FORMAT = {table_format}", {
                "table_format": bp.table_format,
            })

        if bp.integration:
            query.append_nl("INTEGRATION = {integration:i}", {
                "integration": bp.integration,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        query.append_nl("COPY GRANTS")

        return query



from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .ident import SchemaObjectIdent


@dataclass
class StageWithPath:
    stage_name: "SchemaObjectIdent"
    path: str


@dataclass
class StageUploadFile:
    local_path: Path
    stage_path: str



from snowddl.blueprint import WarehouseBlueprint, Ident, AccountObjectIdent
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


warehouse_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "type": {
                "type": "string"
            },
            "size": {
                "type": "string"
            },
            "min_cluster_count": {
                "type": "integer"
            },
            "max_cluster_count": {
                "type": "integer"
            },
            "scaling_policy": {
                "type": "string"
            },
            "auto_suspend": {
                "type": "integer"
            },
            "resource_monitor": {
                "type": "string"
            },
            "global_resource_monitor": {
                "type": "string"
            },
            "enable_query_acceleration": {
                "type": "boolean"
            },
            "query_acceleration_max_scale_factor": {
                "type": "integer"
            },
            "warehouse_params": {
                "type": "object",
                "additionalProperties": {
                    "type": ["boolean", "number", "string"]
                }
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["size"],
        "additionalProperties": False
    }
}


class WarehouseParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'warehouse.yaml', warehouse_json_schema, self.process_warehouse)

    def process_warehouse(self, f: ParsedFile):
        for warehouse_name, warehouse in f.params.items():
            resource_monitor = None

            if warehouse.get('resource_monitor'):
                resource_monitor = AccountObjectIdent(self.env_prefix, warehouse['resource_monitor'])

            if warehouse.get('global_resource_monitor'):
                resource_monitor = Ident(warehouse['global_resource_monitor'])

            bp = WarehouseBlueprint(
                full_name=AccountObjectIdent(self.env_prefix, warehouse_name),
                type=warehouse.get('type', 'STANDARD'),
                size=warehouse['size'],
                auto_suspend=warehouse.get('auto_suspend', 60),
                min_cluster_count=warehouse.get('min_cluster_count'),
                max_cluster_count=warehouse.get('max_cluster_count'),
                scaling_policy=warehouse.get('scaling_policy'),
                resource_monitor=resource_monitor,
                enable_query_acceleration=warehouse.get('enable_query_acceleration'),
                query_acceleration_max_scale_factor=warehouse.get('query_acceleration_max_scale_factor'),
                warehouse_params=self.normalise_params_dict(warehouse.get('warehouse_params', {})),
                comment=warehouse.get('comment'),
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import ResourceMonitorBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class ResourceMonitorResolver(AbstractResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.RESOURCE_MONITOR

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW RESOURCE MONITORS LIKE {env_prefix:ls}", {
            "env_prefix": self.config.env_prefix,
        })

        for r in cur:
            if r['owner'] != self.engine.context.current_role:
                continue

            triggers = {}

            triggers.update(self._parse_triggers(r['notify_at'], 'NOTIFY'))
            triggers.update(self._parse_triggers(r['suspend_at'], 'SUSPEND'))
            triggers.update(self._parse_triggers(r['suspend_immediately_at'], 'SUSPEND_IMMEDIATE'))

            existing_objects[r['name']] = {
                "name": r['name'],
                "credit_quota": int(float(r['credit_quota'])),
                "frequency": r['frequency'],
                "notify_at": r['notify_at'],
                "suspend_at": r['suspend_at'],
                "suspend_immediately_at": r['suspend_immediately_at'],
                "triggers": triggers,
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def _parse_triggers(self, val, action):
        triggers = {}

        if val is None:
            return triggers

        for threshold in val.split(','):
            threshold = threshold.strip(' %')
            triggers[int(threshold)] = action

        return triggers

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ResourceMonitorBlueprint)

    def create_object(self, bp: ResourceMonitorBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE RESOURCE MONITOR {name:i}", {
            "name": bp.full_name,
        })

        query.append_nl("CREDIT_QUOTA = {credit_quota:d}", {
            "credit_quota": bp.credit_quota,
        })

        query.append_nl("FREQUENCY = {frequency}", {
            "frequency": bp.frequency,
        })

        query.append_nl("START_TIMESTAMP = IMMEDIATELY")
        query.append_nl("TRIGGERS")

        for threshold, action in bp.triggers.items():
            query.append_nl("ON {threshold:d} PERCENT DO {action:r}", {
                "threshold": threshold,
                "action": action,
            })

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_resource_monitor)

        return ResolveResult.CREATE

    def compare_object(self, bp: ResourceMonitorBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.credit_quota != row['credit_quota']:
            self.engine.execute_unsafe_ddl("ALTER RESOURCE MONITOR {name:i} SET CREDIT_QUOTA = {credit_quota:d}", {
                "name": bp.full_name,
                "credit_quota": bp.credit_quota,
            }, condition=self.engine.settings.execute_resource_monitor)

            result = ResolveResult.ALTER

        if bp.frequency != row['frequency']:
            self.engine.execute_unsafe_ddl("ALTER RESOURCE MONITOR {name:i} SET FREQUENCY = {frequency} START_TIMESTAMP = IMMEDIATELY", {
                "name": bp.full_name,
                "frequency": bp.frequency,
            }, condition=self.engine.settings.execute_resource_monitor)

            result = ResolveResult.ALTER

        if bp.triggers != row['triggers']:
            query = self.engine.query_builder()

            query.append("ALTER RESOURCE MONITOR {name:i}", {
                "name": bp.full_name,
            })

            query.append_nl("TRIGGERS")

            for threshold, action in bp.triggers.items():
                query.append_nl("ON {threshold:d} PERCENT DO {action:r}", {
                    "threshold": threshold,
                    "action": action,
                })

            self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_resource_monitor)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP RESOURCE MONITOR {name:i}", {
            "name": row['name'],
        }, condition=self.engine.settings.execute_resource_monitor)

        return ResolveResult.DROP



from pathlib import Path
from typing import Dict, Optional

from snowddl.parser.abc_parser import AbstractParser


placeholder_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": ["boolean", "number", "string"]
    }
}


class PlaceholderParser(AbstractParser):
    def load_blueprints(self):
        # This is a special parser that does not load any blueprints, but it loads placeholders instead
        pass

    def load_placeholders(self, placeholder_path: Optional[Path] = None, placeholder_values: Optional[Dict] = None):
        # 1) Start with standard placeholders, always available
        placeholders = {
            'ENV_PREFIX': self.env_prefix,
        }

        # 2) Merge with placeholders from normal config file
        placeholders.update(self.normalise_params_dict(self.parse_single_file(self.base_path / 'placeholder.yaml', placeholder_json_schema)))

        # 3) Merge with placeholders from override config file
        if placeholder_path:
            placeholders.update(self.normalise_params_dict(self.parse_single_file(placeholder_path, placeholder_json_schema)))

        # 4) Merge with explicit placeholder values
        if placeholder_values:
            placeholders.update(self.normalise_params_dict(placeholder_values))

        for name, value in placeholders.items():
            self.config.add_placeholder(name, value)



from re import compile

from snowddl.blueprint import ObjectType
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter, ConvertResult
from snowddl.parser.sequence import sequence_json_schema
from snowddl.parser.table import table_json_schema


cluster_by_syntax_re = compile(r'^(\w+)?\((.*)\)$')
collate_type_syntax_re = compile(r'^(.*) COLLATE \'(.*)\'$')
identity_re = compile(r'^IDENTITY START (\d+) INCREMENT (\d+)$')


class TableConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            # Skip external tables
            if r['is_external'] == 'Y':
                continue

            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "is_transient": r['kind'] == 'TRANSIENT',
                "cluster_by": r['cluster_by'] if r['cluster_by'] else None,
                "change_tracking": bool(r['change_tracking'] == 'ON'),
                "search_optimization": bool(r.get('search_optimization') == 'ON'),
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def dump_object(self, row):
        cols, identities =  self._get_columns(row)

        data = {'columns': cols}

        if row['is_transient']:
            data['is_transient'] = True

        if row['cluster_by']:
            data['cluster_by'] = cluster_by_syntax_re.sub(r'\2', row['cluster_by']).split(', ')

        if row['change_tracking']:
            data['change_tracking'] = True

        if row['search_optimization']:
            data['search_optimization'] = True

        data['comment'] = row['comment']

        data['primary_key'] = self._get_primary_key(row)
        data['unique_keys'] = self._get_unique_keys(row)
        data['foreign_keys'] = self._get_foreign_keys(row)

        object_path = self.base_path / self._normalise_name_with_prefix(row['database']) / self._normalise_name(row['schema']) / 'table'
        object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

        self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, table_json_schema)

        for sequence_name, sequence_data in identities.items():
            object_path = self.base_path / self._normalise_name_with_prefix(row['database']) / self._normalise_name(row['schema']) / 'sequence'
            object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

            self._dump_file(object_path / f"{self._normalise_name(sequence_name)}.yaml", sequence_data, sequence_json_schema)

        return ConvertResult.DUMP

    def _get_columns(self, row):
        cols = {}
        identities = {}

        cur = self.engine.execute_meta("DESC TABLE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        for c in cur:
            m = collate_type_syntax_re.match(c['type'])

            if m:
                col = {
                    "type": m.group(1),
                    "collate": m.group(2),
                }
            else:
                col = {"type": c['type']}

            if c['null?'] == 'N':
                col['type'] = f"{col['type']} NOT NULL"

            if c['default']:
                i = identity_re.match(c['default'])

                if i:
                    sequence_name = self._get_auto_sequence_name(row['name'], c['name'])
                    col['default_sequence'] = sequence_name

                    identities[sequence_name] = {
                        "start": int(i.group(1)),
                        "interval": int(i.group(2)),
                    }
                elif str(c['default']).upper().endswith('.NEXTVAL'):
                    col['default_sequence'] = self._normalise_name_with_prefix(str(c['default'])[:-8])
                else:
                    col['default'] = str(c['default'])

            if c['comment']:
                col['comment'] = c['comment']

            cols[self._normalise_name(c['name'])] = col

        return cols, identities

    def _get_primary_key(self, row):
        constraints = {}

        cur = self.engine.execute_meta("SHOW PRIMARY KEYS IN TABLE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        for r in cur:
            if r['constraint_name'] not in constraints:
                constraints[r['constraint_name']] = {}

            constraints[r['constraint_name']][r['key_sequence']] = r['column_name']

        if not constraints:
            return None

        # It is possible to have multiple PRIMARY KEYS in some cases
        # Return only the first key
        for pk in constraints.values():
            return [self._normalise_name(pk[seq]) for seq in sorted(pk)]

    def _get_unique_keys(self, row):
        constraints = {}
        unique_keys = []

        cur = self.engine.execute_meta("SHOW UNIQUE KEYS IN TABLE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        for r in cur:
            if r['constraint_name'] not in constraints:
                constraints[r['constraint_name']] = {}

            constraints[r['constraint_name']][r['key_sequence']] = r['column_name']

        if not constraints:
            return None

        for uq in constraints.values():
            unique_keys.append([self._normalise_name(uq[seq]) for seq in sorted(uq)])

        return unique_keys

    def _get_foreign_keys(self, row):
        constraints = {}
        foreign_keys = []

        cur = self.engine.execute_meta("SHOW IMPORTED KEYS IN TABLE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        for r in cur:
            if r['fk_name'] not in constraints:
                constraints[r['fk_name']] = {
                    "columns": {},
                    "ref_table": f"{r['pk_database_name']}.{r['pk_schema_name']}.{r['pk_table_name']}",
                    "ref_columns": {},
                }

            constraints[r['fk_name']]['columns'][r['key_sequence']] = r['fk_column_name']
            constraints[r['fk_name']]['ref_columns'][r['key_sequence']] = r['pk_column_name']

        if not constraints:
            return None

        for fk in constraints.values():
            foreign_keys.append({
                "columns": [self._normalise_name(fk['columns'][seq]) for seq in sorted(fk['columns'])],
                "ref_table": self._normalise_name_with_prefix(fk['ref_table']),
                "ref_columns": [self._normalise_name(fk['ref_columns'][seq]) for seq in sorted(fk['ref_columns'])],
            })

        return foreign_keys

    def _get_auto_sequence_name(self, table_name, column_name):
        return self._normalise_name_with_prefix(f"{table_name}_{column_name}_seq")



from abc import ABC, abstractmethod
from enum import Enum
from traceback import format_exc

from concurrent.futures import as_completed
from typing import Dict, TYPE_CHECKING

from snowddl.error import SnowDDLExecuteError, SnowDDLUnsupportedError
from snowddl.blueprint import AbstractBlueprint, DependsOnMixin, Edition, ObjectType

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class ResolveResult(Enum):
    CREATE      = "CREATE"
    ALTER       = "ALTER"
    DROP        = "DROP"
    REPLACE     = "REPLACE"
    SKIP        = "SKIP"
    GRANT       = "GRANT"
    NOCHANGE    = "NOCHANGE"
    ERROR       = "ERROR"
    UNSUPPORTED = "UNSUPPORTED"


class AbstractResolver(ABC):
    skip_on_empty_blueprints = False
    skip_min_edition = Edition.STANDARD

    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine
        self.config = engine.config

        self.object_type = self.get_object_type()
        self.blueprints: Dict[str, AbstractBlueprint] = {}
        self.existing_objects: Dict[str,Dict] = {}

        self.resolved_objects: Dict[str,ResolveResult] = {}
        self.errors: Dict[str,Exception] = {}

    def resolve(self):
        if self._is_skipped():
            return

        self.blueprints = self.get_blueprints()

        try:
            self.existing_objects = self.get_existing_objects()
        except SnowDDLExecuteError as e:
            self.engine.logger.info(f"Could not get existing objects for resolver [{self.__class__.__name__}]: \n{e.verbose_message()}")
            raise e.snow_exc

        self._pre_process()
        self._resolve_create_compare()
        self._resolve_drop()
        self._post_process()

    def destroy(self):
        if self._is_skipped():
            return

        try:
            self.existing_objects = self.get_existing_objects()
        except SnowDDLExecuteError as e:
            self.engine.logger.info(f"Could not get existing objects for resolver [{self.__class__.__name__}]: \n{e.verbose_message()}")
            raise e.snow_exc

        self._pre_process()
        self._destroy_drop()
        self._post_process()

    def _resolve_create_compare(self):
        # Process blueprints in batches sorted in order of dependencies
        for blueprint_names_batch in self._split_blueprints_into_batches():
            tasks = {}

            for full_name in sorted(blueprint_names_batch):
                if full_name in self.existing_objects:
                    tasks[full_name] = (self.compare_object, self.blueprints[full_name], self.existing_objects[full_name])
                else:
                    tasks[full_name] = (self.create_object, self.blueprints[full_name])

            self._process_tasks(tasks)

    def _resolve_drop(self):
        tasks = {}

        for object_full_name in sorted(self.existing_objects):
            # Object exists in blueprints, should not be dropped
            if object_full_name in self.blueprints:
                continue

            tasks[object_full_name] = (self.drop_object, self.existing_objects[object_full_name])

        self._process_tasks(tasks)

    def _destroy_drop(self):
        tasks = {}

        # Drop all existing objects
        for full_name in sorted(self.existing_objects):
            tasks[full_name] = (self.drop_object, self.existing_objects[full_name])

        self._process_tasks(tasks)

    def _process_tasks(self, tasks):
        futures = {}

        for full_name, args in tasks.items():
            futures[self.engine.executor.submit(*args)] = full_name

        for f in as_completed(futures):
            full_name = futures[f]

            try:
                result = f.result()
                if result == ResolveResult.NOCHANGE:
                    self.engine.logger.debug(f"Resolved {self.object_type.name} [{full_name}]: {result.value}")
                else:
                    self.engine.logger.info(f"Resolved {self.object_type.name} [{full_name}]: {result.value}")
            except Exception as e:
                if isinstance(e, SnowDDLUnsupportedError):
                    result = ResolveResult.UNSUPPORTED
                else:
                    result = ResolveResult.ERROR

                if isinstance(e, SnowDDLExecuteError):
                    error_text = e.verbose_message()
                else:
                    error_text = format_exc()

                self.engine.logger.warning(f"Resolved {self.object_type.name} [{full_name}]: {result.value}\n{error_text}")
                self.errors[full_name] = e

            self.resolved_objects[full_name] = result

        self.engine.flush_thread_buffers()

    def _split_blueprints_into_batches(self):
        all_batches = []
        allocated_full_names = set()

        remaining_blueprints = self.blueprints.copy()

        # Create new batches as long as at least one unallocated blueprint remains
        while remaining_blueprints:
            batch = []

            # Allocate blueprints with no dependencies or fully resolved dependencies
            for full_name, bp in remaining_blueprints.items():
                if not isinstance(bp, DependsOnMixin) \
                or not bp.depends_on \
                or all((str(d) in allocated_full_names) for d in bp.depends_on):
                    batch.append(full_name)

            # Allocate blueprints with unresolved dependencies to the last batch
            if len(batch) == 0:
                for full_name in remaining_blueprints:
                    batch.append(full_name)

            # Forget blueprints allocated during this iteration
            for full_name in batch:
                del remaining_blueprints[full_name]

            all_batches.append(batch)
            allocated_full_names.update(batch)

        return all_batches

    def _is_skipped(self):
        if self.engine.context.edition < self.skip_min_edition:
            return True

        if self.object_type in self.engine.settings.exclude_object_types:
            return True

        if self.engine.settings.include_object_types:
            return not (self.object_type in self.engine.settings.include_object_types)

        if self.skip_on_empty_blueprints and not self.get_blueprints():
            return True

        return False

    def _pre_process(self):
        pass

    def _post_process(self):
        pass

    @abstractmethod
    def get_object_type(self) -> ObjectType:
        pass

    @abstractmethod
    def get_blueprints(self) -> Dict[str, AbstractBlueprint]:
        pass

    @abstractmethod
    def get_existing_objects(self) -> Dict[str, Dict]:
        pass

    @abstractmethod
    def create_object(self, bp: AbstractBlueprint) -> ResolveResult:
        pass

    @abstractmethod
    def compare_object(self, bp: AbstractBlueprint, row: Dict) -> ResolveResult:
        pass

    @abstractmethod
    def drop_object(self, row: Dict):
        pass



from enum import Enum
from re import compile, IGNORECASE


class BaseDataType(Enum):
    NUMBER = {
        "base_name": "NUMBER",
        "number_of_properties": 2,
    }

    FLOAT = {
        "base_name": "FLOAT",
        "number_of_properties": 0,
    }

    BINARY = {
        "base_name": "BINARY",
        "number_of_properties": 1,
    }

    BOOLEAN = {
        "base_name": "BOOLEAN",
        "number_of_properties": 0,
    }

    VARCHAR = {
        "base_name": "VARCHAR",
        "number_of_properties": 1,
    }

    DATE = {
        "base_name": "DATE",
        "number_of_properties": 0,
    }

    TIME = {
        "base_name": "TIME",
        "number_of_properties": 1,
    }

    TIMESTAMP_LTZ = {
        "base_name": "TIMESTAMP_LTZ",
        "number_of_properties": 1,
    }

    TIMESTAMP_NTZ = {
        "base_name": "TIMESTAMP_NTZ",
        "number_of_properties": 1,
    }

    TIMESTAMP_TZ = {
        "base_name": "TIMESTAMP_TZ",
        "number_of_properties": 1,
    }

    VARIANT = {
        "base_name": "VARIANT",
        "number_of_properties": 0,
    }

    OBJECT = {
        "base_name": "OBJECT",
        "number_of_properties": 0,
    }

    ARRAY = {
        "base_name": "ARRAY",
        "number_of_properties": 0,
    }

    GEOGRAPHY = {
        "base_name": "GEOGRAPHY",
        "number_of_properties": 0,
    }

    GEOMETRY = {
        "base_name": "GEOMETRY",
        "number_of_properties": 0,
    }

    @property
    def number_of_properties(self) -> int:
        return self.value.get('number_of_properties', 0)

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}>"


class DataType:
    data_type_re = compile(r'^(?P<base_type>[a-z0-9_]+)(\((?P<val1>\d+)(,(?P<val2>\d+))?\))?$', IGNORECASE)

    def __init__(self, data_type_str):
        m = self.data_type_re.match(data_type_str)

        if not m:
            raise ValueError(f"Could not parse data type string [{data_type_str}]")

        try:
            self.base_type = BaseDataType[str(m['base_type']).upper()]
        except KeyError:
            raise ValueError(f"Invalid base data type [{m['base_type']}] in type string [{data_type_str}]")

        self.val1 = int(m['val1']) if self.base_type.number_of_properties >= 1 else None
        self.val2 = int(m['val2']) if self.base_type.number_of_properties >= 2 else None

    def __str__(self):
        if self.base_type.number_of_properties >= 2:
            return f"{self.base_type.name}({self.val1},{self.val2})"
        elif self.base_type.number_of_properties >= 1:
            return f"{self.base_type.name}({self.val1})"

        return self.base_type.name

    def __repr__(self):
        return f"<{self.__class__.__name__}.{str(self)}>"

    def __eq__(self, other):
        if not isinstance(other, DataType):
            raise NotImplemented

        return str(self) == str(other)



from json import loads
from typing import TYPE_CHECKING

from snowddl.blueprint import AccountObjectIdent, Edition

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class SnowDDLContext:
    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine

        cur = self.engine.execute_meta("""
            SELECT CURRENT_ACCOUNT() AS current_account
                , CURRENT_REGION() AS current_region
                , CURRENT_SESSION() AS current_session
                , CURRENT_USER() AS current_user
                , CURRENT_ROLE() AS current_role
                , CURRENT_WAREHOUSE() AS current_warehouse
                , IS_ROLE_IN_SESSION('ACCOUNTADMIN') AS is_account_admin
                , IS_ROLE_IN_SESSION('SYSADMIN') AS is_sys_admin
                , IS_ROLE_IN_SESSION('SECURITYADMIN') AS is_security_admin
                , SYSTEM$BOOTSTRAP_DATA_REQUEST('ACCOUNT') AS bootstrap_account
        """)

        r = cur.fetchone()

        self.current_account = r['CURRENT_ACCOUNT']
        self.current_region = r['CURRENT_REGION']
        self.current_session = r['CURRENT_SESSION']
        self.current_user = r['CURRENT_USER']
        self.current_role = r['CURRENT_ROLE']
        self.current_warehouse = r['CURRENT_WAREHOUSE']

        self.original_role = self.current_role

        self.is_account_admin = r['IS_ACCOUNT_ADMIN']
        self.is_sys_admin = r['IS_SYS_ADMIN']
        self.is_security_admin = r['IS_SECURITY_ADMIN']

        bootstrap_account = loads(r['BOOTSTRAP_ACCOUNT'])

        self.version = bootstrap_account['serverVersion']
        self.edition = Edition[bootstrap_account['accountInfo']['serviceLevelName']]

        self._validate()

    def _validate(self):
        if not self.current_warehouse:
            if self.engine.settings.execute_replace_table:
                raise ValueError("Context error: missing CURRENT_WAREHOUSE in session to apply REPLACE TABLE")

        if not self.is_account_admin:
            if self.engine.settings.execute_account_params:
                raise ValueError("Context error: missing ACCOUNTADMIN role in session to apply ACCOUNT PARAMETERS")

            if self.engine.settings.execute_resource_monitor:
                raise ValueError("Context error: missing ACCOUNTADMIN role in session to apply RESOURCE MONITORS")

    def activate_role_with_prefix(self):
        if not self.engine.config.env_prefix:
            return

        role_with_prefix = AccountObjectIdent(self.engine.config.env_prefix, self.current_role)

        cur = self.engine.execute_meta("SHOW ROLES LIKE {role_with_prefix:lf}", {
            "role_with_prefix": role_with_prefix
        })

        if cur.rowcount == 0:
            self.engine.execute_context_ddl("CREATE ROLE {role_with_prefix:i}", {
                "role_with_prefix": role_with_prefix,
            })

            self.engine.execute_context_ddl("GRANT ROLE {current_role:i} TO ROLE {role_with_prefix:i}", {
                "role_with_prefix": role_with_prefix,
                "current_role": self.current_role,
            })

            self.engine.execute_context_ddl("GRANT ROLE {role_with_prefix:i} TO USER {current_user:i}", {
                "role_with_prefix": role_with_prefix,
                "current_user": self.current_user,
            })

            if not self.is_account_admin:
                self.engine.execute_context_ddl("GRANT ROLE {role_with_prefix:i} TO ROLE ACCOUNTADMIN", {
                    "role_with_prefix": role_with_prefix,
                    "current_role": self.current_role,
                })

        self.engine.execute_meta("USE ROLE {role_with_prefix:i}", {
            "role_with_prefix": role_with_prefix,
        })

        self.current_role = str(role_with_prefix)
        self.engine.flush_thread_buffers()

    def destroy_role_with_prefix(self):
        if not self.engine.config.env_prefix:
            return

        self.engine.execute_meta("USE ROLE {original_role:i}", {
            "original_role": self.original_role,
        })

        self.engine.execute_context_ddl("DROP ROLE {role_with_prefix:i}", {
            "role_with_prefix": self.current_role,
        })

        self.current_role = self.original_role
        self.engine.flush_thread_buffers()



from re import compile, DOTALL

from snowddl.blueprint import ObjectType
from snowddl.converter.abc_schema_object_converter import AbstractSchemaObjectConverter, ConvertResult, YamlLiteralStr
from snowddl.parser.view import view_json_schema


view_text_re = compile(r'^.*\n\)\sas(.*)$', DOTALL)

class ViewConverter(AbstractSchemaObjectConverter):
    def get_object_type(self) -> ObjectType:
        return ObjectType.VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW VIEWS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if r['is_materialized'] == 'true':
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "text": r['text'],
                "is_secure": r['is_secure'] == "true",
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def dump_object(self, row):
        data = {
            "columns": self._get_columns(row),
            "text": YamlLiteralStr(self._get_text(row)),
            "comment": row['comment'],
        }

        object_path = self.base_path / self._normalise_name_with_prefix(row['database']) / self._normalise_name(row['schema']) / 'view'
        object_path.mkdir(mode=0o755, parents=True, exist_ok=True)

        if data:
            self._dump_file(object_path / f"{self._normalise_name(row['name'])}.yaml", data, view_json_schema)
            return ConvertResult.DUMP

        return ConvertResult.EMPTY

    def _get_columns(self, row):
        cols = {}

        cur = self.engine.execute_meta("DESC VIEW {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        for c in cur:
            cols[self._normalise_name(c['name'])] = c['comment'] if c['comment'] else None

        return cols

    def _get_text(self, row):
        # TODO: replace with better implementation when available
        cur = self.engine.execute_meta("SELECT GET_DDL('VIEW', {ident}) AS view_ddl", {
            "ident": f"{row['database']}.{row['schema']}.{row['name']}"
        })

        view_ddl = cur.fetchone()['VIEW_DDL']
        view_text = view_text_re.sub(r'\1', view_ddl).strip(' \n\r\t()')

        # Remove trailing spaces from each line to prevent output formatting issues
        # https://github.com/yaml/pyyaml/issues/411
        view_text = '\n'.join(l.rstrip(' ') for l in view_text.split('\n'))

        return view_text
